/*
 * fbsetfont - program to set font for the Linux framebuffer console
 */

#define _GNU_SOURCE

#include <sys/types.h>
#include <sys/ioctl.h>
#include <stddef.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <error.h>
#include <errno.h>
#include <ctype.h>
#include <getopt.h>

#include <linux/kd.h>

#define VERSION "0.0.2"

/*
 * TODO:
 *
 * - Check all data read from the BDF file.
 */

static void *xmalloc (size_t size);
static void *xmalloc0 (size_t size);
static void *xrealloc (void *mem, size_t size);

#define FONT_MAX_ROWS		32
#define FONT_MAX_ROW_SIZE	4
#define FONT_MAX_CHARS		512

struct char_bitmap {
	unsigned int	 width;
	unsigned int	 height;
	unsigned int	 bytes_per_line;
	unsigned char	 data[FONT_MAX_ROWS * FONT_MAX_ROW_SIZE];
};

static void char_bitmap_init (struct char_bitmap *bitmap,
			      unsigned int width,
			      unsigned int height)
{
	bitmap->width = width;
	bitmap->height = height;
	bitmap->bytes_per_line = (width + 7) >> 3;
	memset (bitmap->data, 0, sizeof (bitmap->data));
}

static void shift_bits_left (unsigned char *bits, unsigned int len)
{
	unsigned int i;

	for (i = 0; i < len - 1; ++i)
		bits[i] = (bits[i] << 1) | ((bits[i+1] >> 7) & 1);
	bits[len - 1] <<= 1;
}

static void shift_bits_right (unsigned char *bits, unsigned int len)
{
	unsigned int i;

	if (len == 0)
		return;
	
	for (i = len - 1; i > 0; --i)
		bits[i] = (bits[i] >> 1) | ((bits[i-1] << 7) & 0x80);
	bits[0] >>= 1;
}

static void char_bitmap_shift_horizontal (struct char_bitmap *bitmap,
					  int shift_amount)
{
	unsigned char *data = bitmap->data;
	unsigned int bpl = bitmap->bytes_per_line;
	unsigned int row;
	int i;

	if (shift_amount > 0) {
		/* right */
		for (row = 0; row < FONT_MAX_ROWS; ++row) {
			for (i = 0; i < shift_amount; ++i)
				shift_bits_right (data, bpl);
			data += bpl;
		}
	} else if (shift_amount < 0) {
		/* left */
		shift_amount = -shift_amount;
		for (row = 0; row < FONT_MAX_ROWS; ++row) {
			for (i = 0; i < shift_amount; ++i)
				shift_bits_left (data, bpl);
			data += bpl;
		}
	}
}

static void char_bitmap_shift_vertical (struct char_bitmap *bitmap,
					int shift_amount)
{
	unsigned char *data = bitmap->data;
	unsigned int bpl = bitmap->bytes_per_line;

	if (shift_amount > 0) {
		/* up */
		memmove (data, data + shift_amount * bpl,
			 (FONT_MAX_ROWS - shift_amount) * bpl);
		memset (data + (FONT_MAX_ROWS - shift_amount) * bpl,
			0, shift_amount * bpl);
	} else if (shift_amount < 0) {
		/* down */
		shift_amount = -shift_amount;
		memmove (data + shift_amount * bpl, data,
			 (FONT_MAX_ROWS - shift_amount) * bpl);
		memset (data, 0, shift_amount * bpl);
	}
}

static void char_bitmap_shift (struct char_bitmap *bitmap,
			       int shift_x, int shift_y)
{
	char_bitmap_shift_horizontal (bitmap, shift_x);
	char_bitmap_shift_vertical (bitmap, shift_y);
}

struct font {
	unsigned int	 width;
	unsigned int	 height;
	unsigned int	 char_count;
	unsigned int	 bytes_per_line;
	unsigned char	*bitmaps;
};

static struct font *font_create (unsigned int width,
				 unsigned int height,
				 unsigned int char_count)
{
	struct font *font = xmalloc (sizeof (struct font));

	font->width = width;
	font->height = height;
	font->char_count = char_count;
	font->bytes_per_line = (width + 7) >> 3;
	font->bitmaps =
		xmalloc0 (font->bytes_per_line * char_count * FONT_MAX_ROWS);

	return font;
}

static void font_free (struct font *font)
{
	if (font) {
		free (font->bitmaps);
		free (font);
	}
}

static unsigned char *font_get_bitmap (struct font *font,
				       unsigned int code)
{
	if (!font)
		return NULL;
	return font->bitmaps + FONT_MAX_ROWS * font->bytes_per_line * code;
}

static int font_put_bitmap (struct font *font,
			    unsigned int code,
			    const struct char_bitmap *bitmap)
{
	unsigned char *dst;

	if (!font || !bitmap)
		return 0;
	if (font->height != bitmap->height
	    || font->width != bitmap->width
	    || font->bytes_per_line != bitmap->bytes_per_line)
		return 0;

	dst = font_get_bitmap (font, code);
	if (!dst)
		return 0;

	memcpy (dst, bitmap->data, font->height * font->bytes_per_line);

	return 1;
}

struct input_file {
	FILE		*file;
	char		*file_name;
	unsigned int	 line;
};

struct input_file *input_file_open (const char *file_name)
{
	struct input_file *file;
	int error_code;

	file = malloc (sizeof (struct input_file));
	if (!file) {
		/*error (0, errno, "no memory for input file '%s'", file_name);*/
		return NULL;
	}

	file->line = 0;
	file->file_name = strdup (file_name);
	if (!file->file_name) {
		error_code = errno;
		free (file);
		errno = error_code;
		return NULL;
	}

	file->file = fopen (file_name, "rt");
	if (!file->file) {
		error_code = errno;
		free (file->file_name);
		free (file);
		errno = error_code;
		return NULL;
	}

	return file;
}

static int input_file_close (struct input_file *file)
{
	int result, error_code;

	if (!file)
		return 0;

	result = fclose (file->file);

	error_code = errno;
	free (file->file_name);
	free (file);
	errno = error_code;

	return result;
}

static int input_file_get_line (struct input_file *file,
				char *buf,
				unsigned int size)
{
	unsigned int len;

	file->line++;
	if (!fgets (buf, size, file->file)) {
		if (feof (file->file))
			return 0;

		error_at_line (0, errno, file->file_name, file->line,
			       "read error");
		return 0;
	}

	len = strlen (buf);
	if (len > 0 && buf[len - 1] == '\n')
		buf[--len] = '\0';
	if (len > 0 && buf[len - 1] == '\r')
		buf[--len] = '\0';

	return 1;
}

struct bdf_header {
	unsigned int	 bb_width;
	unsigned int	 bb_height;
	int		 bb_offset_x;
	int		 bb_offset_y;
	unsigned int	 char_count;
};

#define BDF_BUF_SIZE	256

static inline int bdf_get_line (struct input_file *bdf, char *buf)
{
	while (input_file_get_line (bdf, buf, BDF_BUF_SIZE)) {
		if (*buf != '\0')
			return 1;
		/* skip empty lines */
	}

	return 0;
}

static int first_word_eq (const char *buf, const char *word)
{
	int word_len = strlen (word);
	if (strncmp (buf, word, word_len) == 0) {
		if (buf[word_len] == '\0'
		    || isspace ((unsigned char) buf[word_len]))
			return 1;
	}

	return 0;
}

static int bdf_parse_properties (struct input_file *bdf,
				 char *buf,
				 struct bdf_header *header)
{
	const char *count_str = buf + sizeof("STARTPROPERTIES");
	unsigned int count = strtoul (count_str, NULL, 10);

	(void) header;

	while (bdf_get_line (bdf, buf)) {
		if (count > 0) {
			--count;
			continue;
		}
		if (first_word_eq (buf, "ENDPROPERTIES")) {
			return 1;
		}
		break;
	}

	error_at_line (0, 0, bdf->file_name, bdf->line,
		       "missing ENDPROPERTIES line");

	return 0;
}

static int bdf_parse_font_bbox (struct input_file *bdf,
				const char *buf,
				struct bdf_header *header)
{
	const char *param_str = buf + sizeof("FONTBOUNDINGBOX");
	int bb_width, bb_height, bb_offset_x, bb_offset_y;
	int result;

	result = sscanf (param_str, "%d %d %d %d",
			 &bb_width, &bb_height, &bb_offset_x, &bb_offset_y);
	if (result == 4) {
		header->bb_width = bb_width;
		header->bb_height = bb_height;
		header->bb_offset_x = bb_offset_x;
		header->bb_offset_y = bb_offset_y;
		return 1;
	}

	error_at_line (0, 0, bdf->file_name, bdf->line,
		       "malformed FONTBOUNDINGBOX line");

	return 0;
}

static int bdf_parse_char_count (struct input_file *bdf,
				 const char *buf,
				 struct bdf_header *header)
{
	const char *count_str = buf + sizeof("CHARS");
	char *tail_ptr = NULL;
	unsigned int count;
	
	count = strtoul (count_str, &tail_ptr, 10);
	if (tail_ptr &&
	    (*tail_ptr == '\0' || isspace ((unsigned char) *tail_ptr))) {
		header->char_count = count;
		return 1;
	}

	error_at_line (0, 0, bdf->file_name, bdf->line,
		       "malformed CHARS line");

	return 0;
}

static int bdf_parse_header (struct input_file *bdf,
			     struct bdf_header *header)
{
	char buf[BDF_BUF_SIZE];
	int have_font_bbox = 0;
	int have_chars = 0;

	while (1) {
		if (!bdf_get_line (bdf, buf)) {
			error_at_line
				(0, 0, bdf->file_name, bdf->line,
				 "end of file while parsing BDF header");
			return 0;
		}

		if (first_word_eq (buf, "FONTBOUNDINGBOX")) {
			if (!bdf_parse_font_bbox (bdf, buf, header))
				return 0;
			have_font_bbox = 1;
			continue;
		}
		if (first_word_eq (buf, "STARTPROPERTIES")) {
			if (!bdf_parse_properties (bdf, buf, header))
				return 0;
			continue;
		}
		if (first_word_eq (buf, "CHARS")) {
			if (!bdf_parse_char_count (bdf, buf, header))
				return 0;
			have_chars = 1;
			break;
		}
	}

	if (!have_font_bbox) {
		error_at_line (0, 0, bdf->file_name, bdf->line,
			       "FONTBOUNDINGBOX line not found");
		return 0;
	}
	if (!have_chars) {
		error_at_line (0, 0, bdf->file_name, bdf->line,
			       "CHARS line not found");
		return 0;
	}

	return 1;
}

struct uint_set {
	unsigned int	 size;
	unsigned int	*data;
};

#if 0
static void uint_set_init (struct uint_set *set)
{
	set->size = 0;
	set->data = NULL;
}
#endif

static void uint_set_clear (struct uint_set *set)
{
	if (set) {
		set->size = 0;
		free (set->data);
		set->data = NULL;
	}
}

static void uint_set_add (struct uint_set *set, unsigned int value)
{
	unsigned int i;

	if (!set->data) {
		set->size = 1;
		set->data = xmalloc (sizeof (unsigned int));
		set->data[0] = value;
		return;
	}

	for (i = 0; i < set->size; ++i) {
		if (set->data[i] == value)
			return;
		else if (set->data[i] > value)
			break;
	}

	set->size++;
	set->data = xrealloc (set->data, sizeof (unsigned int) * set->size);
	if (i + 1 < set->size)
		memmove (&set->data[i+1], &set->data[i],
			 sizeof (unsigned int) * (set->size - i - 1));
	set->data[i] = value;
}

static inline int uint_set_contains (const struct uint_set *set,
				     unsigned int value)
{
	unsigned int i;

	if (!set)
		return 0;

	for (i = 0; i < set->size; ++i) {
		if (set->data[i] == value)
			return 1;
		else if (set->data[i] > value)
			return 0;
	}

	return 0;
}

struct font_params {
	unsigned int	 char_count;
	int		 unicode_mode;
	struct uint_set	 unicode_map[FONT_MAX_CHARS];
};

static void font_params_init (struct font_params *params)
{
	memset (params, 0, sizeof (struct font_params));
}

static void font_params_clear (struct font_params *params)
{
	unsigned int i;

	if (!params)
		return;

	for (i = 0; i < FONT_MAX_CHARS; ++i)
		uint_set_clear (&params->unicode_map[i]);
}

static struct font *bdf_create_font (struct bdf_header *header,
				     struct font_params *params)
{
	struct font *font;

	font = font_create (header->bb_width, header->bb_height,
			    params->char_count);

	return font;
}

struct bdf_char_attrs {
	unsigned int	 encoding;
	unsigned int	 dwidth;
	unsigned int	 bb_width;
	unsigned int 	 bb_height;
	int		 bb_offset_x;
	int		 bb_offset_y;
};

static int bdf_parse_char_encoding (struct input_file *bdf,
				    char *buf,
				    struct bdf_char_attrs *attrs)
{
	const char *param_str = buf + sizeof("ENCODING");
	int enc1, enc2;

	if (sscanf (param_str, "%d", &enc1) == 1) {
		if (enc1 >= 0) {
			attrs->encoding = enc1;
			return 1;
		}
		if (sscanf (param_str, "%d %d", &enc1, &enc2) == 2) {
			attrs->encoding = enc2;
			return 1;
		}
		attrs->encoding = (unsigned int) -1;
		return 1;
	}

	error_at_line (0, 0, bdf->file_name, bdf->line,
		       "malformed ENCODING line");

	return 0;
}

static int bdf_parse_char_dwidth (struct input_file *bdf,
				  char *buf,
				  struct bdf_char_attrs *attrs)
{
	const char *param_str = buf + sizeof("DWIDTH");
	int dwidth_1, dwidth_2;

	if (sscanf (param_str, "%d %d", &dwidth_1, &dwidth_2) == 2) {
		if (dwidth_2 != 0) {
			error_at_line (0, 0, bdf->file_name, bdf->line,
				       "DWIDTH second parameter is non-zero, "
				       "this is not supported");
			return 0;
		}
		attrs->dwidth = dwidth_1;
		return 1;
	}

	error_at_line (0, 0, bdf->file_name, bdf->line,
		       "malformed DWIDTH line");

	return 0;
}

static int bdf_parse_char_bbx (struct input_file *bdf,
			       char *buf,
			       struct bdf_char_attrs *attrs)
{
	const char *param_str = buf + sizeof("BBX");
	int bb_width, bb_height, bb_offset_x, bb_offset_y;
	int result;

	result = sscanf (param_str, "%d %d %d %d",
			 &bb_width, &bb_height, &bb_offset_x, &bb_offset_y);
	if (result == 4) {
		attrs->bb_width = bb_width;
		attrs->bb_height = bb_height;
		attrs->bb_offset_x = bb_offset_x;
		attrs->bb_offset_y = bb_offset_y;
		return 1;
	}

	error_at_line (0, 0, bdf->file_name, bdf->line,
		       "malformed BBX line");

	return 0;
}

static int bdf_parse_char_attrs (struct input_file *bdf,
				 char *buf,
				 struct bdf_char_attrs *attrs)
{
	int have_dwidth = 0;
	int have_bbx = 0;

	attrs->encoding = (unsigned int) -1;
	while (1) {
		if (!bdf_get_line (bdf, buf)) {
			error_at_line (0, 0, bdf->file_name, bdf->line,
				       "end of file while parsing BDF "
				       "character attributes");
			return 0;
		}

		if (first_word_eq (buf, "ENCODING")) {
			if (!bdf_parse_char_encoding (bdf, buf, attrs))
				return 0;
			continue;
		}
		if (first_word_eq (buf, "DWIDTH")) {
			if (!bdf_parse_char_dwidth (bdf, buf, attrs))
				return 0;
			have_dwidth = 1;
			continue;
		}
		if (first_word_eq (buf, "BBX")) {
			if (!bdf_parse_char_bbx (bdf, buf, attrs))
				return 0;
			have_bbx = 1;
			continue;
		}

		if (first_word_eq (buf, "BITMAP"))
			break;
	}

	if (!have_dwidth) {
		error_at_line (0, 0, bdf->file_name, bdf->line,
			       "DWIDTH line not found");
		return 0;
	}
	if (!have_bbx) {
		error_at_line (0, 0, bdf->file_name, bdf->line,
			       "BBX line not found");
		return 0;
	}

	return 1;
}

static int get_hex_byte (const char *str)
{
	unsigned int value;

	if (!isxdigit ((unsigned char) str[0])
	    || !isxdigit ((unsigned char) str[1]))
		return -1;

	if (sscanf (str, "%02x", &value) != 1)
		return -1;

	return value;
}

static int bdf_parse_bitmap_line (struct input_file *bdf,
				  char *buf,
				  struct bdf_char_attrs *attrs,
				  unsigned char *bits)
{
	unsigned int i;
	unsigned int byte_count = (attrs->bb_width + 7) >> 3;

	for (i = 0; i < byte_count; ++i) {
		int byte = get_hex_byte (buf);
		if (byte == -1) {
			error_at_line (0, 0, bdf->file_name, bdf->line,
				       "malformed bitmap data");
			return 0;
		}
		*bits++ = byte;
		buf += 2;
	}

	return 1;
}

static int bdf_parse_char_bitmap (struct input_file *bdf,
				  char *buf,
				  struct bdf_header *header,
				  struct bdf_char_attrs *attrs,
				  struct char_bitmap *bitmap)
{
	unsigned int i;
	unsigned char *bits = bitmap->data;
	unsigned int bpl = bitmap->bytes_per_line;

	for (i = 0; i < attrs->bb_height; ++i) {
		if (!bdf_get_line (bdf, buf)) {
			error_at_line (0, 0, bdf->file_name, bdf->line,
				       "end of file while parsing BDF "
				       "character bitmap");
			return 0;
		}
		if (!bdf_parse_bitmap_line (bdf, buf, attrs, bits))
			return 0;
		bits += bpl;
	}

	/* Align bitmap with the global font bbox */
	char_bitmap_shift ( bitmap,
			    attrs->bb_offset_x - header->bb_offset_x,
			    ( (attrs->bb_offset_y + attrs->bb_height)
			      - (header->bb_offset_y + header->bb_height) ) );

	return 1;
}

static int bdf_put_char_bitmap (struct input_file *bdf,
				struct bdf_header *header,
				struct font *font,
				struct font_params *params,
				struct bdf_char_attrs *attrs,
				struct char_bitmap *bitmap)
{
	unsigned int encoding = attrs->encoding;
	unsigned int i;

	/* ignore unencoded chars */
	if (encoding == (unsigned int) -1)
		return 1;

	if (params->unicode_mode) {
		for (i = 0; i < font->char_count; ++i)
			if (uint_set_contains (&params->unicode_map[i],
					       encoding)) {
				font_put_bitmap (font, i, bitmap);
			}
	} else {
		if (encoding < font->char_count)
			font_put_bitmap (font, encoding, bitmap);
	}

	return 1;
}

static int bdf_parse_char (struct input_file *bdf,
			   struct bdf_header *header,
			   struct font *font,
			   struct font_params *params)
{
	struct char_bitmap bitmap;
	struct bdf_char_attrs attrs;
	char buf[BDF_BUF_SIZE];

	char_bitmap_init (&bitmap, font->width, font->height);

	if (!bdf_get_line (bdf, buf)) {
		error_at_line (0, 0, bdf->file_name, bdf->line,
			       "premature end of file while parsing "
			       "BDF character");
		return 0;
	}
	if (!first_word_eq (buf, "STARTCHAR")) {
		error_at_line (0, 0, bdf->file_name, bdf->line,
			       "STARTCHAR expected");
		return 0;
	}

	memset (&attrs, 0, sizeof attrs);
	if (!bdf_parse_char_attrs (bdf, buf, &attrs))
		return 0;

	if (!bdf_parse_char_bitmap (bdf, buf, header, &attrs, &bitmap))
		return 0;

	if (!bdf_get_line (bdf, buf)) {
		error_at_line (0, 0, bdf->file_name, bdf->line,
			       "premature end of file while parsing "
			       "BDF character");
		return 0;
	}
	if (!first_word_eq (buf, "ENDCHAR")) {
		error_at_line (0, 0, bdf->file_name, bdf->line,
			       "ENDCHAR expected");
		return 0;
	}

	if (!bdf_put_char_bitmap (bdf, header, font, params, &attrs, &bitmap))
		return 0;

	return 1;
}

static struct font *bdf_parse_file (struct input_file *bdf,
				    struct font_params *params)
{
	struct bdf_header header;
	struct font *font;
	unsigned int i;

	memset (&header, 0, sizeof header);

	if (!bdf_parse_header (bdf, &header))
		return NULL;

	font = bdf_create_font (&header, params);

	for (i = 0; i < header.char_count; ++i) {
		if (!bdf_parse_char (bdf, &header, font, params)) {
			font_free (font);
			return NULL;
		}
	}

	return font;
}

static struct font *bdf_parse (const char *file_name,
			       struct font_params *params)
{
	struct input_file *bdf;
	struct font *result;

	bdf = input_file_open (file_name);
	if (!bdf) {
		error (0, errno, "cannot open '%s' for reading", file_name);
		return NULL;
	}

	result = bdf_parse_file (bdf, params);

	input_file_close (bdf);

	return result;
}


static int set_font (struct font *font)
{
	struct console_font_op font_op;

	memset (&font_op, 0, sizeof font_op);
	font_op.op = KD_FONT_OP_SET;
	font_op.flags = 0;
	font_op.width = font->width;
	font_op.height = font->height;
	font_op.charcount = font->char_count;
	font_op.data = font->bitmaps;

	if (ioctl (1, KDFONTOP, &font_op) == -1) {
		error (0, errno, "error setting font");
		return 0;
	}

	return 1;
}


static int set_font_from_file (const char *file_name,
			       struct font_params *params)
{
	struct font *font;
	int result;

	font = bdf_parse (file_name, params);
	if (!font)
		return 0;

	result = set_font (font);

	font_free (font);

	return result;
}

static int sfm_add_unicode (struct font_params *params,
			    unsigned int glyph_number,
			    unsigned int unicode)
{
	uint_set_add (&params->unicode_map[glyph_number], unicode);

	return 1;
}

static int sfm_get_number (const char **buf_ptr, unsigned int *value_ptr)
{
	const char *buf = *buf_ptr;
	char *tail_ptr = NULL;
	unsigned long value;

	value = strtoul (buf, &tail_ptr, 0);
	if (!tail_ptr
	    || (*tail_ptr != '\0' && !isspace ((unsigned char) *tail_ptr)))
		return 0;

	*value_ptr = value;

	while (isspace ((unsigned char) *tail_ptr))
		++tail_ptr;
	*buf_ptr = tail_ptr;

	return 1;
}

static int sfm_get_unicode (const char **buf_ptr, unsigned int *value_ptr)
{
	const char *ptr = *buf_ptr;
	char tmp_buf[5];

	if (ptr[0] != 'u' && ptr[0] != 'U')
		return 0;
	if (ptr[1] != '+')
		return 0;
	if (!isxdigit ((unsigned char) ptr[2])
	    || !isxdigit ((unsigned char) ptr[3])
	    || !isxdigit ((unsigned char) ptr[4])
	    || !isxdigit ((unsigned char) ptr[5]))
		return 0;
	memcpy (tmp_buf, ptr + 2, 4);
	tmp_buf[4] = '\0';
	*value_ptr = strtoul (tmp_buf, NULL, 16);

	ptr += 6;
	while (isspace ((unsigned char) *ptr))
		++ptr;
	*buf_ptr = ptr;

	return 1;
}

static int sfm_parse_line (struct input_file *sfm_file,
			   const char *buf,
			   struct font_params *params)
{
	unsigned int glyph_number, unicode;

	while (*buf && isspace ((unsigned char )*buf))
		++buf;
	if (*buf == '\0' || *buf == '#')
		return 1;

	if (!sfm_get_number (&buf, &glyph_number)) {
		error_at_line (0, 0, sfm_file->file_name, sfm_file->line,
			       "malformed glyph number");
		return 0;
	}
	if (glyph_number >= params->char_count) {
		error_at_line (0, 0, sfm_file->file_name, sfm_file->line,
			       "glyph number too big");
		return 0;
	}

	while (*buf && (*buf != '#')) {
		if (!sfm_get_unicode (&buf, &unicode)) {
			error_at_line (0, 0, sfm_file->file_name,
				       sfm_file->line,
				       "malformed unicode number");
			return 0;
		}
		if (!sfm_add_unicode (params, glyph_number, unicode))
			return 0;
	}

	return 1;
}

static int sfm_parse (struct input_file *sfm_file,
		      struct font_params *params)
{
	char buf[1024];

	while (input_file_get_line (sfm_file, buf, sizeof buf)) {
		if (!sfm_parse_line (sfm_file, buf, params))
			return 0;
	}

	return 1;
}

static int sfm_load_from_file (struct font_params *params,
			       const char *file_name)
{
	struct input_file *sfm_file;
	int result;

	sfm_file = input_file_open (file_name);
	if (!sfm_file) {
		error (0, errno, "cannot open '%s' for reading", file_name);
		return 0;
	}

	result = sfm_parse (sfm_file, params);

	input_file_close (sfm_file);

	return result;
}

static void show_help (void)
{
	puts ("usage: fbsetfont [OPTION]... BDF-FILE");
	puts ("Set font for the Linux framebuffer console.\n");
	puts ("  -c, --char-count=N    set character count (256 or 512)");
	puts ("  -u, --sfm=SFM-FILE    specify the Screen Font Map file");
	puts ("  -h, --help            display this help and exit");
	puts ("  -V, --version         display version information and exit\n");
	puts ("WARNING: current version does not really set the SFM "
	      "in kernel!");
}

static void show_version (void)
{
	puts ("fbsetfont " VERSION);
}


static const struct option options[] = {
	{ "help",	no_argument,		NULL,	'h' },
	{ "version",	no_argument,		NULL,	'V' },
	{ "char-count",	required_argument,	NULL,	'c' },
	{ "sfm",	required_argument,	NULL,	'u' },
	{ NULL,		0,			NULL,	0   }
};


int main (int argc, char *argv[])
{
	struct font_params params;
	int int_value;

	font_params_init(&params);
	params.char_count = 256;
	params.unicode_mode = 0;

	while (1) {
		int option_index = 0;
		int c = getopt_long (argc, argv, "hVc:u:",
				     options, &option_index);

		/* detect the end of the options */
		if (c == -1)
			break;

		switch (c) {

		  case 0:
			/* do nothing for options which set a flag */
			if (options[option_index].flag)
				break;
			/* no flag and no tag - that's a bug */
			abort ();

		  case 'h':	/* -h, --help */
			show_help ();
			exit (EXIT_SUCCESS);

		  case 'V':	/* -V, --version */
			show_version ();
			exit (EXIT_SUCCESS);

		  case 'c':	/* -c, --char-count=N */
			int_value = atoi (optarg);
			if (int_value != 256 && int_value != 512)
				error (EXIT_FAILURE, 0,
				       "char count can be either 256 or 512");
			params.char_count = int_value;
			break;

		  case 'u':	/* -u, --sfm=SFM-FILE */
			if (!sfm_load_from_file (&params, optarg))
				exit (EXIT_FAILURE);
			params.unicode_mode = 1;
			break;

		  case '?': 	/* invalid option or missing argument */
			/* getopt_long already printed an error message */
			exit (EXIT_FAILURE);

		  default:	/* no handler for the option - bug! */
			abort ();
		}
	}

	if (optind < argc) {
		if (optind != argc - 1)
			error (EXIT_FAILURE, 0,
			       "too many parameters");

		if (!set_font_from_file (argv[optind], &params))
			exit (EXIT_FAILURE);
	} else {
		error (EXIT_FAILURE, 0, "missing font argument");
	}

	font_params_clear (&params);

	return 0;
}



void *xmalloc (size_t size)
{
	void *mem = malloc (size);
	if (!mem)
		error (EXIT_FAILURE, errno, "memory exhausted");
	return mem;
}

void *xmalloc0 (size_t size)
{
	void *mem = calloc (size, 1);
	if (!mem)
		error (EXIT_FAILURE, errno, "memory exhausted");
	return mem;
}

void *xrealloc (void *mem, size_t size)
{
	void *new_mem = realloc (mem, size);
	if (!new_mem)
		error (EXIT_FAILURE, errno, "memory exhausted");
	return new_mem;
}

/* vim: set sw=8 cino=(0\:2=6l0: */
